#include<sys/types.h>
#include<pthread.h>
#include<sys/time.h>
#include<string.h>
#include<stdio.h>
#include<stdlib.h>

/* Start address to write in Flash */
#define bl_start_address 0x0
/* warning: bl_frame_length must be greater than 4 */
#define bl_frame_length 5
u_int8_t bl_frame1[bl_frame_length];
u_int8_t bl_frame2[bl_frame_length];

u_int8_t *bl_frame_data, *bl_frame_current;
/* Security code length must be the same as a frame length */
u_int8_t bl_security_code[] = {'c','r','o','l'};

u_int8_t bl_flags;
u_int8_t bl_frame;

/* max number of frame error before givevup */
#define bl_timeout_limit 3

#define bl_flag_frame   &bl_flags,0
#define bl_flag_prog    &bl_flags,1
#define bl_flag_error   &bl_flags,2

/* Acknowledge message for the client */
#define bl_ack  'z'
#define bl_nack 'n'
#define bl_err  'e'

/* Client variables - To be moved later!!! */
u_int8_t * bl_client_data_ptr;

/* Buffer to simulate communication */
u_int8_t bl_data_buffer;
u_int8_t bl_return_buffer;

int mode_prog()
{
  return 1;
}

int bl_hit = 0;

int kbhit()
{
  if(bl_hit)
  {
    bl_hit = 0;
    return 1;
  }
  else
    return 0;
}

void bit_clear(u_int8_t * data, int bit)
{
  *data = *data & ~(1<<bit);
}

void bit_set(u_int8_t * data, int bit)
{
  *data = *data | (1<<bit);
}

char bit_test(u_int8_t * data, int bit)
{
  if(*data & (1<<bit))
    return 1;
  else
    return 0;
}

char bl_check_crc(u_int8_t * data, unsigned int length)
{
  unsigned int i;
  u_int8_t crc=0;

  for(i=0; i<length-1;i++)
    crc+=data[i];

  for(i=0; i<length-1;i++)
    printf("%c.",data[i]);
  printf("\r\n");

  printf("Received crc 0x%x calc: 0x%x\r\n",data[i],crc);  

  if(data[i] == crc)
    return 1;
  else
    return 0;
}

u_int8_t bl_calc_crc(u_int8_t * data, unsigned int length)
{
  unsigned int i;
  u_int8_t crc=0;

  for(i=0; i<length;i++)
    crc+=data[i];

  for(i=0; i<length;i++)
    printf("%c.",data[i]);
  printf("\r\n");

  return crc;
}

u_int8_t bl_get_ack()
{
  u_int8_t buf;

  buf = bl_return_buffer;
  bl_return_buffer = 0;
  return buf;
}

void bl_send_ack()
{
  bl_return_buffer = bl_ack;
}

void bl_send_nack()
{
  bl_return_buffer = bl_nack;
}

void bl_send_error()
{
  bl_return_buffer = bl_err;
}

int bl_prog_flash(u_int16_t address, u_int8_t data)
{
  printf("%c at 0x%x\r\n",data,address);

  return 0;
}

/*******************************************************************/
void * bl_client_task(void * arg)
{
  int frame_len = 0;
  struct timeval lastcommand, now, diff;
  struct timezone here;
  enum {bl_state_end, bl_state_waitack, bl_state_data, bl_state_crc, bl_state_resend} bl_client_state;

  u_int8_t data[] = "cool\t\0\0\0Salut\n";
  u_int8_t framedata[bl_frame_length-1];
  u_int8_t ack;

  bl_client_state    = bl_state_data;
  bl_client_data_ptr = data;
  gettimeofday(&lastcommand,&here);

  while(bl_client_state != bl_state_end)
  { 
    gettimeofday(&now,&here);
    timersub(&now,&lastcommand,&diff);

    if(diff.tv_sec >= 1)
    {

      switch(bl_client_state)
      {
	case bl_state_data:
	  /* Send a data byte */
	  printf("char sent: %c\r\n",*bl_client_data_ptr);
	  bl_data_buffer       = *bl_client_data_ptr;
	  framedata[frame_len] = *bl_client_data_ptr;

	  bl_client_data_ptr++;
	  if(*bl_client_data_ptr == '\n')
	    bl_client_data_ptr = data;

	  /* Update status */
	  bl_hit = 1;
	  gettimeofday(&lastcommand,&here);
	  frame_len++;

	  /* check if a frame is completed */
	  if( frame_len == bl_frame_length-1 )
	    bl_client_state = bl_state_crc;
	  break;

	case bl_state_waitack:
	  /* Wait for a ack for the target */
	  ack = bl_get_ack();
	  switch(ack)
	  {
	    case bl_ack:
	      bl_client_state = bl_state_data;
	      break;
	    case bl_nack:
	      bl_client_state = bl_state_resend;
	      break;
	    case bl_err:
	      bl_client_state = bl_state_end;
	      break;
	  }

	  /* Update status */
	  gettimeofday(&lastcommand,&here);

	  break;

	case bl_state_resend:
	  /* Resend a frame after a nack */
	  printf("char sent: %c\r\n",framedata[frame_len]);
	  bl_data_buffer = framedata[frame_len];

	  /* Update status */
	  bl_hit = 1;
	  gettimeofday(&lastcommand,&here);
	  frame_len++;

	  /* check if a frame is completed */
	  if( frame_len == bl_frame_length-1 )
	    bl_client_state = bl_state_crc;
	  break;

	case bl_state_crc:
	  /* Send a crc for the current frame */
	  printf("Sending crc 0x%x\r\n",bl_calc_crc(framedata,4));
	  bl_data_buffer = bl_calc_crc(framedata,4);

	  /* Update status */
	  bl_hit = 1;
	  frame_len = 0;
	  gettimeofday(&lastcommand,&here);

	  bl_client_state = bl_state_waitack;
	  break;
      }
    }
  }
  printf("Program End\r\n");
}

char bl_getc()
{
  return bl_data_buffer;
}

/*******************************************************************/
void rs232_rec_int() {
  u_int8_t * frame_tmp;
  int i;

  while(kbhit())
  {
    bl_frame_current[bl_frame] = bl_getc();
    bl_frame++;

    if(bl_frame == bl_frame_length)
    {
      bit_set(bl_flag_frame);
      frame_tmp        = bl_frame_data;
      bl_frame_data    = bl_frame_current;
      bl_frame_current = frame_tmp;

      bl_frame = 0;
    }
  }
}

void * bl_rs232_task(void * arg)
{
  
  struct timeval lastcommand, now, diff;
  struct timezone here;

  gettimeofday(&lastcommand,&here);

  for(;;)
  {    
    gettimeofday(&now,&here);
    timersub(&now,&lastcommand,&diff);

    if(diff.tv_usec >= 50000)
    {
      gettimeofday(&lastcommand,&here);
      rs232_rec_int();  
    }
  }
}

/*******************************************************************/
void * bl_prog_task(void * arg)
{
  enum {bl_state_security_check, bl_state_length, bl_state_programming} bl_state_prog;
  
  u_int32_t bl_prog_length;
  u_int32_t bl_data_received = 0;
  u_int16_t bl_address = bl_start_address;
  u_int8_t  bl_timeout = 0;

  int i;

  u_int8_t * bl_char_ptr = (u_int8_t*)&bl_prog_length;

  /* Init the fame counter */
  bl_frame         = 0;
  bl_frame_current = bl_frame1;
  bl_frame_data    = bl_frame2;

  /* Check that the programming mode is selected */  
  if(!mode_prog())
    exit(0);

  bit_set(bl_flag_prog);
  bit_clear(bl_flag_error);
  bl_state_prog = bl_state_security_check;

  /* Programming main loop */
  while(bit_test(bl_flag_prog))
  {
    /* Check if a correct frame has been received */
    if(bit_test(bl_flag_frame))
    {
      printf("frame received: ");
      for(i=0; i<bl_frame_length; i++)
	printf("%c",bl_frame_data[i]);
      printf("\r\n");

      /*Clear the received frame */
      bit_clear(bl_flag_frame);

      if(bl_check_crc(bl_frame_data,bl_frame_length))
	/* Process the frame */
	switch(bl_state_prog)
	{
	  /* Check security code */
	  case bl_state_security_check:
	    printf("Security check\r\n");
	    for(i=0; i<bl_frame_length-1; i++)
	      if(bl_frame_data[i] != bl_security_code[i])
		bit_set(bl_flag_error);
	    
	    if(!bit_test(bl_flag_error))
		bl_state_prog = bl_state_length;
	    break;

	    /* Get the program lengh */
	    /* First 4 bytes of the frame data, rest of the frame is ignored */
	  case bl_state_length :
	    for(i=0; i<4; i++)
	      bl_char_ptr[i] = bl_frame_data[i];

	    bl_state_prog = bl_state_programming;
	    printf("prog length: 0x%lx\r\n",bl_prog_length);
	    break;

	    /* Program the flash */
	  case bl_state_programming:
	    i = 0;

	    while(bl_data_received < bl_prog_length && i < bl_frame_length-1)
	    {
	      if(bl_prog_flash(bl_address, bl_frame_data[i]))
		bit_set(bl_flag_error);
	      i++;
	      bl_data_received++;
	    }

	    printf("%d/%d done\r\n", bl_data_received, bl_prog_length);
	    if(bl_data_received >= bl_prog_length)
	      bit_clear(bl_flag_prog); 
	    break;
	}
      else
	bit_set(bl_flag_error);
      

      /* Check for errors in the current frame */
      if(bit_test(bl_flag_error))
      {
	printf("Frame error\r\n");
	/* Error with this frame */
	bl_timeout++;
	bit_clear(bl_flag_error);

	if(bl_timeout > bl_timeout_limit)
	{
	  printf("Giveup\r\n");
	  /* Giveup programming */
	  bit_clear(bl_flag_prog);
	  /* Send ERR to the host */
	  bl_send_error();
	}
	else
	  /* Send a NACK to the host, the frame should be resent */
	  bl_send_nack();
      }
      else
      {
	/* Send ACK to the host */
	bl_send_ack();
      }
    }  
  }
}


int main()
{
  pthread_t prog_task;
  pthread_t client_task;
  pthread_t rs232_task;
  int rc;

  printf("ucBootloader Program\r\n");

  rc = pthread_create(&rs232_task, NULL, bl_rs232_task, NULL);
  if (rc){
    printf("ERROR; return code from pthread_create() is %d\n", rc);
    exit(-1);
  }
  rc = pthread_create(&prog_task, NULL, bl_prog_task, NULL);
  if (rc){
    printf("ERROR; return code from pthread_create() is %d\n", rc);
    exit(-1);
  }
  rc = pthread_create(&client_task, NULL, bl_client_task, NULL);
  if (rc){
    printf("ERROR; return code from pthread_create() is %d\n", rc);
    exit(-1);
  }

  while(1);
}
